# -*- coding: utf-8 -*-

# Copyright 2023 Ubiratan S. Freitas


# This file is part of Lattesh.
# 
# Lattesh is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by the 
# Free Software Foundation, either version 3 of the License, or (at your option) 
# any later version.
# 
# Lattesh is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
# more details.
# 
# You should have received a copy of the GNU General Public License along with 
# Lattice Tools. If not, see <https://www.gnu.org/licenses/>. 



bl_info = {
    "name": "Lattice Tools",
    "author": "Ubiratan S. Freitas",
    "blender": (2, 80, 0),
    "description": "Tools for generating lattice structures",
    "category": "Add Mesh",
}






if "bpy" in locals():
    import importlib
    importlib.reload(operators)
    importlib.reload(ui)
else:
    import bpy
    from bpy.types import PropertyGroup
    from bpy.props import (
        BoolProperty,
        FloatProperty,
        EnumProperty,
        IntVectorProperty,
        PointerProperty,
    )

    import os
    if hasattr(os, 'add_dll_directory'):
        import sys
        dll_path = os.path.join(sys.prefix, 'Library/bin')
        dpath = os.add_dll_directory(dll_path)
    from . import (
        operators,
        ui,
    )


cell_items = operators.cell_items

class SceneProperties(PropertyGroup):
    r_ratio: FloatProperty(
        name="R_ratio",
        description="Ratio of actual radius to metaball radius",
        default=0.8, 
        min=0.01,
        max=0.99,
    )

    radius: FloatProperty(
        name="Radius",
        description="Actual radius of resulting mesh",
        default=0.5, 
        min=1.0e-6,
        max=20.0,
        unit='LENGTH',
    )

    cellsize: FloatProperty(
        name="Cell size",
        description="Cell size",
        default=10, 
        min=1.0e-2,
        unit='LENGTH',
    )

    nelems: IntVectorProperty(
        name="Nr. cells",
        description="Number of cells in each direction",
        default=(5, 5, 3), 
        min=1,
        size=3,
    )

    celltype: EnumProperty(
        items=cell_items,
        name="Cell type",
        description="Cell type",
        default=cell_items[0][0],
    )

    connect_boundary: BoolProperty(
        name="Connect Boundary",
        description="Connect Boundary",
        default=False,
    )

    charac_length: FloatProperty(
        name="Caracteristic length",
        description="Caracteristic length",
        default=10, 
        min=1.0e-2,
        unit='LENGTH',
    )

    lin_deflect: FloatProperty(
        name="Linear deflection",
        description="Maximum linear deflection for tesselation",
        default=0.1, 
        min=1.0e-6,
        max=2.0,
        unit='LENGTH',
    )

    ang_deflect: FloatProperty(
        name="Angular deflection",
        description="Maximum angular deflection for tesselation",
        default=1.5707963267948966,  # 90 degrees in radians
        min=1.0e-3,
        max=2.0943951023931953, # 120 degrees in radians
        subtype='ANGLE',
        unit='ROTATION',
    )




classes = [
    SceneProperties,
    ui.VIEW3D_PT_lattices_generate,
    ui.VIEW3D_PT_lattice_metaball,
    ui.VIEW3D_PT_lattice_OCC,
    operators.ObjectOCCLattice,
    operators.ObjectMetaLattice,
    operators.ObjectCreateMeshLattice,
    operators.ObjectFillMeshLattice,
]

if operators.have_scipy:
    classes.append(operators.ObjectFillVoronoiLattice)


def menu_func(self, context):
    for cl in classes:
        if cl is not SceneProperties:
            self.layout.operator(cl.bl_idname)




def register():
    for cls in classes:
        bpy.utils.register_class(cls)
    bpy.types.Scene.lattice_mesh = PointerProperty(type=SceneProperties)
    bpy.types.VIEW3D_MT_object.append(menu_func)




def unregister():
    for cls in classes:
        bpy.utils.unregister_class(cls)
    del bpy.types.Scene.lattice_mesh
    bpy.types.VIEW3D_MT_object.remove(menu_func)

if __name__ == "__main__":
    register()



