# -*- coding: utf-8 -*-
# Copyright 2023 Ubiratan S. Freitas


# This file is part of Lattesh.
# 
# Lattesh is free software: you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by the 
# Free Software Foundation, either version 3 of the License, or (at your option) 
# any later version.
# 
# Lattesh is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
# more details.
# 
# You should have received a copy of the GNU General Public License along with 
# Lattice Tools. If not, see <https://www.gnu.org/licenses/>. 



import bpy, bmesh
from mathutils import Vector, Quaternion, Matrix
from mathutils.bvhtree import BVHTree
import numpy as np
import math
import os
from glob import glob

try:
    from scipy.spatial import Voronoi, Delaunay
    have_scipy = True
except ModuleNotFoundError:
    have_scipy = False





# Opencascade imports
try:
    from OCC.Core.BRep import BRep_Tool
    from OCC.Core.BRepPrimAPI import BRepPrimAPI_MakeSphere, BRepPrimAPI_MakeCylinder
    from OCC.Core.BRepAlgoAPI import BRepAlgoAPI_Fuse
    from OCC.Core.BRepBuilderAPI import BRepBuilderAPI_Transform
    from OCC.Core.BRepMesh import BRepMesh_IncrementalMesh
    from OCC.Core.TopExp import TopExp_Explorer
    from OCC.Core.TopoDS import topods_Face
    from OCC.Core.TopTools import TopTools_ListOfShape
    from OCC.Core.BRepAdaptor import BRepAdaptor_Surface
    from OCC.Core.GeomAbs import GeomAbs_Plane
    from OCC.Core.TopAbs import TopAbs_FACE
    from OCC.Core.TopLoc import TopLoc_Location
    from OCC.Core.gp import gp_Pnt, gp_Trsf, gp_Ax2, gp_Ax1, gp_Dir
    have_occ = True
except ModuleNotFoundError:
    have_occ = False







class Cell:
    def __init__(self, verts=None, edges=None):
        if verts is None:
            self.verts = np.zeros((0,3), dtype=np.float32)
            self.nverts = 0
        else:
            self.verts = verts
            self.nverts = verts.shape[0]
        if edges is None:
            self.edges = np.zeros((0,2), dtype=np.int32)
            self.nedges = 0
        else:
            self.edges = edges
            self.nedges = edges.shape[0]

        self.check()

    def check(self):
        if self.nverts == 0:
            return
        if len(self.verts.shape) != 2 or self.verts.shape[1] != 3:
            raise ValueError("verts should be an N x 3 array")

        if len(self.edges.shape) != 2 or self.edges.shape[1] != 2:
            raise ValueError("edges should be an M x 2 array")

        if self.verts.min() < 0.0 or self.verts.max() > 1.0:
            raise ValueError("All vertices' coordinates must be in [0.0, 1.0]")

        if self.edges.min() < 0 or self.edges.max() > self.nverts:
            raise ValueError("Edge(s) referencing inexistent vertex")

        if set(range(self.nverts)) != set(self.edges.flat):
            raise ValueError("Each vertex must be part of an edge")

    def from_file(self, fname):
        name = ''
        verts = []
        edges = []
        with open(fname) as f:
            for line in f:
                stripped = line.strip()
                if len(stripped) > 0 and stripped[0] != '#':
                    if len(name) == 0:
                        name = stripped
                    else:
                        sp = stripped.split(',')
                        if len(sp) == 3:
                            verts.append(list(map(float, sp)))
                        elif len(sp) == 2:
                            edges.append(list(map(int, sp)))
        self.verts = np.array(verts, dtype=np.float32)
        self.nverts = len(verts)
        self.edges = np.array(edges, dtype=np.int32)
        self.nedges = len(edges)
        self.check()
        self.name = name



fpath = os.path.dirname(os.path.abspath(__file__)) 

Cells = {}
cell_items = []
cell_files = glob(os.path.join(fpath, 'cells/*.cell'))
cell_files.sort()
for k, cf in enumerate(cell_files):
    c = Cell()
    c.from_file(cf)
    Cells[c.name] = c
    cell_items.append((c.name, c.name, c.name, k))

def triangulate_verts(bm, max_len):
    """Makes a Delaunay triangulation of selected vertices, delete edges longer than max_len"""
    verts = [v for v in bm.verts if v.select]
    vnp = np.zeros((len(verts),3), dtype=np.float32)
    for k, v in enumerate(verts):
        vnp[k,:] = np.array(v.co, dtype=np.float32)
    tri = Delaunay(vnp)
    eds_s = set()
    for r in tri.simplices:
        for k in range(3):
            a = r[k]
            b = r[k + 1]
            if a < b:
                eds_s.add((a, b))
            else:
                eds_s.add((b, a))
    for e in eds_s:
        if (verts[e[0]].co - verts[e[1]].co).length < max_len:
            new_edge = bm.edges.new((verts[e[0]], verts[e[1]]))
            new_edge.select = True


def make_npvoronoi(carac_length, dims, corner):
    """Makes a Voronoi lattice"""
    
    carac3 = carac_length ** 3
    padding = 2
    
    dims = dims + (2 * padding * carac_length)
    
    npoints = math.ceil(dims.prod() / carac3)
    
    npoints = max(npoints, 30) # create at least this amount of points
    
    print('voronoi npoints:', npoints)
    
    p = np.random.random((npoints,3))
    p = p * dims.reshape((1,3)) + (corner.reshape((1,3)) - padding * carac_length)


    vor = Voronoi(p)

    verts = vor.vertices

    # mark voronoi vertices outside extended bounding box
    lowerlims = corner.reshape((1,3)) - padding * carac_length
    upperlims = lowerlims + dims
    vertok = np.logical_and(np.all(verts > lowerlims, axis=1), np.all(verts < upperlims, axis=1))

    # new vertices
    new_verts = verts[vertok,:].copy()

    # build a table translating old indexes to new_verts
    okidx = np.zeros((len(verts),), dtype=np.int32)
    okidx[np.logical_not(vertok)] = -1
    newidx = okidx.cumsum() + np.arange(len(verts), dtype=np.int32)

    eds_s = set()

    for r in vor.ridge_vertices:
        if not -1 in r:
            for k in range(len(r) - 1):
                a = r[k]
                b = r[k + 1]
                if vertok[a] and vertok[b]:
                    na = newidx[a]
                    nb = newidx[b]
                    if na < nb:
                        eds_s.add((na,nb))
                    else:
                        eds_s.add((nb,na))

    edges = np.array([list(x) for x in eds_s], dtype=np.int32)
    return new_verts, edges



def make_nplattice(cell, cell_size, nelems, corner):
    """Constructs a lattice grid"""
    ncells = nelems.prod()
    verts = np.zeros((ncells * cell.nverts, 3), dtype=np.float32)
    edges = np.zeros((ncells * cell.nedges, 2), dtype=np.int32)
    cellcorner = np.zeros((1,3), dtype=np.float32)
    edgespos = np.ones((1,2), dtype=np.int32)

    # Generate a lattice with copied cells of unit size
    for i in range(nelems[0]):
        for j in range(nelems[1]):
            for k in range(nelems[2]):
                cellcorner[0, 0] = i
                cellcorner[0, 1] = j
                cellcorner[0, 2] = k
                baseidx = i + (j * nelems[0]) + (nelems[0] * nelems[1] * k)
                vertpos = baseidx * cell.nverts
                cellidx = baseidx * cell.nedges
                verts[vertpos : (vertpos + cell.nverts), :] = cellcorner + cell.verts
                edges[cellidx: (cellidx + cell.nedges), :] = cell.edges + edgespos * vertpos

    # Scale the whole lattice and translate it to corner
    verts = verts * cell_size + corner

    return verts, edges

def nplattice2meshdata(v, e):
    verts = [tuple(v[i,:]) for i in range(v.shape[0]) ]
    edges = [tuple(e[i,:]) for i in range(e.shape[0]) ]
    return verts, edges

def make_meshlattice(name, verts, edges, cellsize):
    mesh = bpy.data.meshes.new(name+'Mesh')
    mesh.from_pydata(verts, edges, [])
    
    bm = bmesh.new()
    
    bm.from_mesh(mesh)
    bmesh.ops.remove_doubles(bm, verts=bm.verts[:], dist=1.0e-3*cellsize)
    return bm, mesh


def vor_intersect(vor, obj, depsgraph):
    """Clips a Voronoi 3D lattice to object
    """

    tree = BVHTree.FromObject(obj, depsgraph)
    disttol = 5.0e-5
    analized_edges = set()
    for rg, rv in enumerate(vor.ridge_vertices):
        intersec = 0
        for k in range(len(rv) - 1):
            v0 = min((rv[k], rv[k + 1]))
            v1 = max((rv[k], rv[k + 1]))
            if v1 < 0 or (v0, v1) in analized_edges:
                continue
            p0 = vor.vertices[v0]
            p1 = vor.vertices[v1]
            d = p1 - p0
            dis = np.linalg.norm(d)
            loc, n, idx, hdis = tree.ray_cast(p0, d, dis)
            if hdis is not None:
                loc2, n2, idx2, hdis2 = tree.ray_cast(p1, -d, dis)
                if (loc - loc2).length > disttol / 2.0:
                    double_edges.add(e)
                elif hdis > disttol and abs(dis - hdis) > disttol:
                    r = hdis/dis
                    cedges.append((e, r))
                


def intersect(bm, obj, depsgraph):
    """Clips mesh lattice to object
    
    Returns a mesh where all vertices and edges are inside or on the surface
    of object. Edges that intercept the object are clipped.
    """
    tree = BVHTree.FromObject(obj, depsgraph)
    outside = set()
    inside = set()
    valencies = set()
    surfverts = set()
    cedges = []
    disttol = 5.0e-5
    double_edges = set()
    
    # Find the edges that cross object
    for e in bm.edges:
        p0 = e.verts[0].co
        p1 = e.verts[1].co
        d = p1 - p0
        dis = d.length
        loc, n, idx, hdis = tree.ray_cast(p0, d, dis)
        if hdis is not None:
            # check for multiple crossings
            loc2, n2, idx2, hdis2 = tree.ray_cast(p1, -d, dis)
            if loc2 and (loc - loc2).length > disttol / 2.0:
                double_edges.add(e)
            elif hdis > disttol and abs(dis - hdis) > disttol:
                r = hdis/dis
                cedges.append((e, r))

    print('edges', len(cedges))
    print('double edges', len(double_edges))

    # Delete double crossing edges
    if double_edges:
        bmesh.ops.delete(bm, geom=list(double_edges), context='EDGES')


    dper = dict()
    for k in cedges:
        dper[k[0]] = k[1]
    
    tmp = np.array(list(dper.values()))
    print('min dist', tmp.min(), 'max dist', tmp.max())

    # Split the found edges on crossing point
    bmesh.ops.bisect_edges(bm, edges=[ e[0] for e in cedges ], cuts=1, edge_percents=dper)['geom_split']

    # Find the vertices that are on the surface and identify their outside and inside edges based on 
    # surface normal
    for v in bm.verts:
        loc, n, idx, hdis = tree.find_nearest(v.co, 1.1*disttol)
        if loc is not None:
            surfverts.add(v)
            valencies.add(len(v.link_edges))
            for ed in v.link_edges:
                p1 = ed.other_vert(v)
                u_v = p1.co - v.co
                u_v = u_v / u_v.length * 0.2
                pn = v.co + u_v
                an1 = n.angle(p1.co - v.co)
                loc2, n2, idx, hdis = tree.find_nearest(p1.co, 1.1* ed.calc_length())
                if hdis and hdis > 0:
                    an2 = n2.angle((p1.co - loc2))
                else:
                    an2 = math.pi
                if  hdis and hdis > disttol and (an1 < math.pi / 2 + 1e-3 or an2 < math.pi / 2 + 1e-3) :
                    outside.add(ed)
                else:
                    inside.add(ed)

    print('Found', len(surfverts), 'surfverts')

    print('Valencies:', valencies)

    print('inside', len(inside), 'outside', len(outside))

    invalid_out = set([v for v in outside if not v.is_valid])
    outside = outside - invalid_out

    invalid_in = set([v for v in inside if not v.is_valid])
    inside = inside - invalid_in
    
    print('inv out', len(invalid_out), 'inv in', len(invalid_in), 'in out', len(inside.intersection(outside)))

    # Delete the outside edges from on-the-surface vertices
    bmesh.ops.delete(bm, geom=list(outside), context='EDGES')

    # Remove invalid surfverts
    invalid = set([v for v in surfverts if not v.is_valid])
    print('Found {} invalid surfverts'.format(len(invalid)))
    surfverts = surfverts - invalid
    
    # Find all vertices linked to the ones on the surface
    to_process = surfverts.copy()
    processed = set()
    while to_process:
        v = to_process.pop()
        for e in v.link_edges:
            other_v = e.other_vert(v)
            if other_v not in processed:
                to_process.add(other_v)
        processed.add(v)

    # Delete gemometry not  linked to surface vertices
    to_delete = set(bm.verts) - processed
    print(len(to_delete), 'to be deleted and', len(processed), 'internal vertices found')

    if len(to_delete) < len(bm.verts):
        # do not delete all vertices
        bmesh.ops.delete(bm, geom=list(to_delete), context='VERTS')


    # Leave only surface vertices selected
    bm.select_mode = {'VERT'}

    for v in bm.verts:
        v.select = False

    for v in surfverts:
        v.select = True

    bm.select_flush_mode()

def test_ring(na, nb):
    twopi = 2 * math.pi
    ra = np.random.rand(na).cumsum()
    rb = np.random.rand(nb).cumsum()
    tha = ra / ra[-1] * twopi - math.pi
    thb = rb / rb[-1] * twopi - math.pi
    verts = np.zeros((na + nb, 3), dtype=np.float32)
    verts[:na,:] = np.stack((np.cos(tha), np.sin(tha), np.zeros(na)), axis=1)
    verts[na:,:] = np.stack((np.cos(thb), np.sin(thb), np.ones(nb)*2), axis=1)
    faces, masq = ring_connect_faces(tha, thb)
    faces[np.logical_not(masq)] += na
    faces = list(faces)
    bottom = list(range(na))
    bottom.reverse()
    top = list(range(na,na+nb))
    faces += [bottom, top]
    mesh = bpy.data.meshes.new('RingMesh')
    mesh.from_pydata(verts, [], faces)
    newobj  = bpy.data.objects.new('Ring', mesh)

    # Link object to scene
    bpy.context.scene.collection.objects.link(newobj)
    

    

def angle_dist(a, b):
    d = abs(b-a)
    return min(d, 2 * math.pi - d)

def ring_connect3d(ringa, ringb):
    na, nb = len(ringa), len(ringb)
    distab = np.zeros((na, nb))
    ran_a = np.arange(na)
    ran_b = np.arange(nb)
    rankab = np.zeros((na, nb), dtype=np.int32)
    rankba = np.zeros((nb, na), dtype=np.int32)
    for k in range(na):
        distab[k, :] = np.linalg.norm(ringb - ringa[k, :], axis=1)
    for k in range(na):
        rankab[k, distab[k, :].argsort()] = ran_b
    for k in range(nb):
        rankba[k, distab[:, k].argsort()] = ran_a
    idx = distab.argmin()
    posa, posb = np.unravel_index(idx, distab.shape)
    nposa = (posa + 1) % na
    nposb = (posb + 1) % nb
    faces = np.zeros((na + nb, 3), dtype=np.int32)

    ntris = 0
    ntrisa = 0
    ntrisb = 0
    while ntrisa < na and ntrisb < nb:
        rnb = rankba[nposb, posa]
        rna = rankab[nposa, posb]
        if rnb < rna:
            make_b = True
        elif rna < rnb:
            make_b = False
        elif distab[posa, nposb] <= distab[nposa, posb]:
        #if distab[posa, nposb] <= distab[nposa, posb]:
            make_b = True
        else:
            make_b = False

        if make_b:
            faces[ntris, :] = [posa, nposb + na, posb + na]
            posb = nposb
            nposb = (posb + 1) % nb
            ntrisb += 1
        else:
            faces[ntris, :] = [posa, nposa , posb + na]
            posa = nposa
            nposa = (posa + 1) % na
            ntrisa += 1
        ntris += 1
    while ntrisa < na:
        faces[ntris, :] = [posa, nposa , posb + na]
        posa = nposa
        nposa = (posa + 1) % na
        ntrisa += 1
        ntris += 1
    while ntrisb < nb:
        faces[ntris, :] = [posa, nposb + na, posb + na]
        posb = nposb
        nposb = (posb + 1) % nb
        ntrisb += 1
        ntris += 1
    return faces

def ring_connect_faces(ang_a, ang_b):
    twopi = 2 * math.pi
    na = len(ang_a)
    nb = len(ang_b)
    faces = np.zeros((na+nb, 3), dtype=np.int32)
    face_masq = np.zeros(faces.shape, dtype=bool)
    saidx = ang_a.argsort()
    sbidx = ang_b.argsort()
    sa = ang_a[saidx]
    sb = ang_b[sbidx]
    
    posa = 0
    nposa = 1
    dists = np.zeros((nb,2), dtype=ang_b.dtype)
    dists[:,0] = np.abs(sb - sa[0])
    dists[:,1] = twopi - dists[:,0]
    posb = dists.min(axis=1).argmin()
    nposb = (posb + 1) % nb
    ntris = 0
    ntrisa = 0
    ntrisb = 0
    
    while ntrisa < na and ntrisb < nb:
        dista = angle_dist(sa[posa], sb[nposb])
        distb = angle_dist(sb[posb], sa[nposa])
        if dista <= distb:
            faces[ntris,:] = [saidx[posa], sbidx[nposb], sbidx[posb]]
            face_masq[ntris,:] = [True, False, False]
            posb = (posb + 1) % nb
            nposb = (posb + 1) % nb
            ntrisb += 1
        else:
            faces[ntris,:] = [saidx[posa], saidx[nposa], sbidx[posb]]
            face_masq[ntris,:] = [True, True, False]
            posa = (posa + 1) % na
            nposa = (posa + 1) % na
            ntrisa += 1
        ntris += 1
    while ntrisa < na:
        faces[ntris,:] = [saidx[posa], saidx[nposa], sbidx[posb]]
        face_masq[ntris,:] = [True, True, False]
        posa = (posa + 1) % na
        nposa = (posa + 1) % na
        ntrisa += 1
        ntris += 1
    while ntrisb < nb:
        faces[ntris,:] = [saidx[posa], sbidx[nposb], sbidx[posb]]
        face_masq[ntris,:] = [True, False, False]
        posb = (posb + 1) % nb
        nposb = (posb + 1) % nb
        ntrisb += 1
        ntris += 1
        
    return faces, face_masq

vecz = np.array([0, 0, 1], dtype=np.float32)
eye = np.eye(3)
inv_eye = eye.copy()
inv_eye[1,1] = -1.0
inv_eye[2,2] = -1.0
twopi = math.pi * 2
min_crossnorm = 1e-3

def make_rotmatrix(x):
    """Make a rotation matrix that transform the given unity vector to the z axis"""
    c_angle = x @ vecz
    raxis = np.cross(x, vecz)
    if np.linalg.norm(raxis) < min_crossnorm:
        if c_angle > 0.0:
            rotmat = eye
        elif c_angle < 0.0:
            rotmat = inv_eye
    else:
        rotmat = np.array(Matrix.Rotation(math.acos(c_angle), 3, raxis))
    return rotmat
    

def aux_show_node():
    nodes = [[1, 0, 0],
             [0, 1, 0],
             [0, 0, 1],
             [0,  0.7071067811865475,  0.7071067811865475]
             ]
    nodes = np.array(nodes)
    verts, tris, nsides, shape = occ_compute_node(nodes)
    
    print('Nverts:', len(verts), 'ntris:', len(tris))
    print('Tris:')
    print(tris)
    name = 'OCC Node'
    mesh = bpy.data.meshes.new(name+'Mesh')
    mesh.from_pydata(verts, [], list(tris))
    newobj  = bpy.data.objects.new(name, mesh)

    # Link object to scene
    bpy.context.scene.collection.objects.link(newobj)


def boolean_union(shape_a, shape_b):
    lo1 = TopTools_ListOfShape()
    lo1.Append(shape_a)
    lo2 = TopTools_ListOfShape()
    lo2.Append(shape_b)
    
    op = BRepAlgoAPI_Fuse()
    op.SetArguments(lo1)
    op.SetTools(lo2)

    op.SetCheckInverted(False)
    #op.SetFuzzyValue(r/50)
    #op.SetUseOBB(True)
    op.Build()
    return op.Shape()


def occ_compute_node(ev, r=0.5, mesh_pars=(0.1, False, math.radians(60), False), shuffle=False):
    node_origin = gp_Ax2(gp_Pnt(0, 0, 0), gp_Dir(0, 0, 1))
    shape = BRepPrimAPI_MakeSphere(node_origin, r).Shape()
    vecz = np.array([0.0, 0.0, 1.0])
    rotate = gp_Trsf()
    nedges = len(ev)
    rotmats = np.zeros((2, 3, nedges))
    if shuffle:
        seq = np.random.permutation(nedges)
    else:
        seq = list(range(nedges))

    for k in seq:
        v = ev[k, :]
        c_angle = v @ vecz
        raxis = np.cross(vecz, v)
        if np.linalg.norm(raxis) < min_crossnorm:
            if c_angle > 0.0:
                rotdir = gp_Dir(0.0, 0.0, 1.0)
            elif c_angle < 0.0:
                rotdir = gp_Dir(1.0, 0.0, 0.0)
        else:
            rotdir = gp_Dir(*raxis)

        rotate.SetRotation(gp_Ax1(node_origin.Location(), rotdir), math.acos(c_angle))
        for row in range(2):
            for col in range(3):
                rotmats[row, col, k] = rotate.Value(col + 1, row + 1)
        try:
            if nedges == 1:
                cyl = BRepPrimAPI_MakeCylinder(node_origin, r, 1.1 * r).Shape()
                cyl1 = BRepBuilderAPI_Transform(cyl, rotate, True).Shape()
                shape = boolean_union(shape, cyl1)
            else:
                c_angles = ev @ v
                s_arg = c_angles.argsort()
                minangle = math.acos(c_angles[s_arg[-2]])
                cyl_len = max(1.1 * r, 1.1 * r /math.tan(minangle/2))
                cyl = BRepPrimAPI_MakeCylinder(node_origin, r, cyl_len ).Shape()
                cyl1 = BRepBuilderAPI_Transform(cyl, rotate, True).Shape()
                #op = BRepAlgoAPI_Fuse(shape, cyl1)
                newshape = boolean_union(shape, cyl1)
                while newshape is None:
                    ang_change = + np.random.randn() / 20
                    print(f'Boolean fuse failed. Trying angle change of {math.degrees(ang_change):.2f} degrees')
                    rotate.SetRotation(gp_Ax1(node_origin.Location(), rotdir), math.acos(c_angle) + ang_change)
                    cyl1 = BRepBuilderAPI_Transform(cyl, rotate, True).Shape()
                    newshape = boolean_union(shape, cyl1)
                shape = newshape
        except:
            print('Error in OCC. Type of shape:', type(shape))
            print(shape)
            print(f'k: {k}')
            print(ev)
            raise

    #
    # Mesh the shape
    #
    BRepMesh_IncrementalMesh(shape, *mesh_pars)

    bt = BRep_Tool()
    ex = TopExp_Explorer(shape, TopAbs_FACE)
    verts = []
    uniq_verts = []
    vertsd = {}
    plane_verts = []
    tris = []
    plane_ind = []
    ind_tri = np.array([2, 0, 1], dtype=np.int32)
    while ex.More():
        face = topods_Face(ex.Current())
        location = TopLoc_Location()
        facing = (bt.Triangulation(face, location))
        if facing is not None:
            local_index = []
            for k in range(1, facing.NbNodes() + 1):
                pnt = facing.Node(k)
                point = list(pnt.Coord())
                dpoint = tuple((np.array(point) * 100).round().astype(np.int32))
                verts.append(point)
                if dpoint not in vertsd:
                    vertsd[dpoint] = len(uniq_verts)
                    uniq_verts.append(point)
                local_index.append(vertsd[dpoint])
    
            surf = BRepAdaptor_Surface(face, True)
            surf_type = surf.GetType()
            local_index = np.array(local_index)
            if surf_type == GeomAbs_Plane:
                face_plane = surf.Plane()
                face_normal = np.array(list(face_plane.Axis().Direction().Coord()))
                c_angles = ev @ face_normal
                idx = c_angles.argmax()
                if idx in plane_ind:
                    fidx = plane_ind.index(idx)
                    plane_verts[fidx] = np.unique(np.concatenate((plane_verts[fidx], local_index )))
                else:
                    plane_ind.append(idx)
                    plane_verts.append(local_index)
            else:
                for i in range(1, facing.NbTriangles()+1):
                    trian = facing.Triangle(i)
                    np_tri = np.array(list(trian.Get()), dtype=np.int32)
                    np_tri = np.array(local_index[np_tri - 1])
                    if (np_tri - np_tri[ind_tri]).prod() != 0 :
                        tris.append(np_tri)
        ex.Next()

    uniq_verts = np.array(uniq_verts)
    tris = np.array(tris, dtype=np.int32)
    ed_verts = []
    if plane_verts:
        nsides = len(plane_verts[0])
        all_planeverts = set()
        for pl in plane_verts:
            all_planeverts.update(pl)
        plane_tris = set()
        plane_ind = np.array(plane_ind)
        edverts = []
        for k in plane_ind.argsort():
            pl = plane_verts[k]
            plset = set(pl)
            edvert = set()
            for k in range(len(tris)):
                for x in tris[k]:
                    if x in plset:
                        plane_tris.add(k)
                        edvert.update(set(tris[k]) - all_planeverts)
                        break
            edverts.append(list(edvert))
        verts = np.array(verts)
        # Remove tris that contain a vertex in plane_verts (using set operations)
        tris = tris[list(set(range(len(tris))) - plane_tris), :]

        nverts = len(uniq_verts)
        edge_verts = set()
        non_planeedge_verts = set(range(nverts)) -all_planeverts

        # Look for tris with the same verts
        repeated_tris = set()
        tri_dict = {}
        verts_to_check = set()
        for k, tr in enumerate(tris):
            tr_b = tr.copy()
            tr_b.sort()
            key = tuple(tr_b)
            plist = tri_dict.setdefault(key, [])
            plist.append(k)
            if len(plist) > 1:
                repeated_tris.update(plist)
                verts_to_check.update(key)
        if repeated_tris:
            tris = tris[list(set(range(len(tris))) - repeated_tris), :]
            verts_to_remove = verts_to_check - set(tris.ravel())
            non_planeedge_verts -= verts_to_remove
            print(f'Found {len(repeated_tris)} repeated tris and removed {len(verts_to_remove)} verts.')
            
        for x in edverts:
            non_planeedge_verts -= set(x)
            edge_verts.update(x)
        edge_verts = np.array(list(edge_verts), dtype=np.int32)
        non_planeedge_verts = np.array(list(non_planeedge_verts), dtype=np.int32)
        vert_order = np.concatenate((edge_verts, non_planeedge_verts))
        uniq_verts = uniq_verts[vert_order, :]
        vert_ind = -np.ones(nverts, dtype=np.int32)
        vert_ind[vert_order] = np.arange(len(vert_order), dtype=np.int32)
        tris = vert_ind[tris]
        for x in edverts:
            ed_verts.append(vert_ind[x])
    else:
        nsides = 0


#    print('Total vertices')
#    print(len(verts))
#    print('Unique verts')
#    print(len(uniq_verts))
#    print('Plane faces found:', len(plane_verts))
#    print('Plane nb verts:', plane_nb_verts.min(), 'to', plane_nb_verts.max())
#    print('Non plane verts:', len(uniq_verts) - plane_nb_verts.sum())
#    print('Nb of tris:', len(tris))

    return uniq_verts, tris, nsides, ed_verts, shape



def createMesh(target, r=0.5, lin_deflect=0.1, ang_deflect=math.radians(60.)):
    """Gives a volume to each edge in target using cylinders of radius r and nsides sides"""
    min_angle = math.radians(30)
    min_length = r / math.tan(min_angle / 2.0)
    quantization = 100
    tmesh = target.data
    name = target.name + "Cyl"
    nverts = len(tmesh.vertices)
    nedges = len(tmesh.edges)
    
    verts = np.empty(nverts * 3, dtype=np.float32)
    edges = np.empty(nedges * 2, dtype=np.int32)

    tmesh.vertices.foreach_get('co', verts)
    tmesh.edges.foreach_get('vertices', edges)
    
    verts.shape = (nverts,3)
    edges.shape = (nedges,2)
    
    vec_edges = verts[edges[:,1],:] - verts[edges[:,0],:]
    edge_len = np.linalg.norm(vec_edges, axis=1, keepdims=True)
    vec_edges /= edge_len
    
    
    # Quantize edge orientations
    qedges = (vec_edges * quantization).astype(np.int32)
    
    dqedges = {}
    for k in range(nedges):
        qv = tuple(qedges[k,:])
        dqedges.setdefault(qv,[]).append(k)

    edge_t = np.zeros(nedges, dtype=np.int32)
    edge_types = []
    for k, t in enumerate(dqedges.keys()):
        edge_types.append(t)
        edge_t[dqedges[t]] = k


    mesh = bpy.data.meshes.new(name+'Mesh')

    print(f'{nedges} total edges, {len(dqedges)} quantized orientations, {min_length:.2f}, minimal edge length.')

    
    # Generate all possible edge orientations, including reverse
    tempset = set()
    for k in dqedges.keys():
        tempset.add(k)
        bw = -np.array(list(k))
        tempset.add(tuple(bw))

    dvecs = {}
    listdir = list(tempset)
    for k, v in enumerate(listdir):
        dvecs[v] = k

    # Record all edges and their directions w.r.t. each node (vertex)
    # nodes is a list containing, for every vertex, a list of pairs [ind, dir],
    # where ind is the index of an edge that is connected with the vertex and
    # dir is true if the edge starts in this vertex and false otherwise
    nodes = [None] * nverts
    for k in range(nedges):
        v0, v1 = edges[k,0], edges[k,1]
        t = (k, True)
        if nodes[v0] is None:
            nodes[v0] = [t]
        else:
            nodes[v0].append(t)
        t = (k, False)
        if nodes[v1] is None:
            nodes[v1] = [t]
        else:
            nodes[v1].append(t)

    # Classify all nodes based on their (discretized) edge orientations
    node_types = {}
    single_nodes = set()
    discarded_edges = set()
    short_edges = set()
    for k in range(nverts):
        if nodes[k] is None:
            # Ignore isolated vertices
            continue
        if len(nodes[k]) == 1:
            single_nodes.add(k)
            if edge_len[nodes[k][0][0]] < min_length:
                # Discard single nodes with small edge length
                discarded_edges.add(nodes[k][0])
                continue
        cand_edges = []
        cand_dir = []
        for kk, eddir in nodes[k]:
            if kk in discarded_edges:
                continue
            if edge_len[kk] < min_length:
                if len(nodes[edges[kk, int(eddir)]]) == 1:
                    # Discard single nodes with small edge length
                    discarded_edges.add(kk)
                    continue
                else:
                    # Mark short edges between non single nodes
                    short_edges.add(kk)
            cand_edges.append(kk)
            cand_dir.append(eddir)
        if len(cand_edges) == 0:
            continue
        node_edges = np.array(cand_edges, dtype=np.int32) # list of edge indeces
        edge_dir = np.array(cand_dir, dtype=np.int32)
        edge_vec = qedges[node_edges, :].copy()
        edge_vec[np.flatnonzero(np.logical_not(edge_dir)),:] *= -1 # flip the edge vector so that all point away from vertex
        dir_keya = np.array([dvecs[tuple(x)] for x in edge_vec]) # retrieve the index of each quantitized vector in listdir
        sd = dir_keya.argsort()
        dir_key = tuple(dir_keya[sd])
        node_types.setdefault(dir_key, []).append(k)
        temp = np.stack((node_edges, edge_dir.astype(np.int32)), axis=1)
        nodes[k] = temp[sd,:] # sort the node's list of edges according to edge position in dict's keys

    print(nverts, 'total vertices,', len(node_types), 'distinct nodes,', len(single_nodes), 'single nodes,', len(discarded_edges), 'discarded edges and', len(short_edges), 'short edges.')



    total_verts = 0
    total_faces = 0
    processed_nodes = []
    occ_nsides = None
    nr_nodetypes = len(node_types)
    for k, (nt, nodelist) in enumerate(node_types.items()):
        if nr_nodetypes > 100:
            if k % 10 == 0:
                print(f'Processing node {k} of {nr_nodetypes}', end='\r')
        # Reconstruct edge vectors based on quantitized vectors
        ev = np.array([listdir[d] for d in nt], dtype=np.float32) / quantization

        # Compute node geomery with Opencascade
        try:
            retry_count = 0
            node_verts, node_faces, node_nsides, node_edverts, shape = occ_compute_node(ev, r, mesh_pars=(lin_deflect, False, ang_deflect, False))
            while len(node_edverts) != len(nt) and retry_count < 20:
                retry_count += 1
                print(f'Wrong number of edge verts in processed node {k} with {len(nt)} edges and {len(node_edverts)} found edges. Retry nb. {retry_count}.', end='\r')
                node_verts, node_faces, node_nsides, node_edverts, shape = occ_compute_node(ev, r, mesh_pars=(lin_deflect, False, ang_deflect, False), shuffle=True)
            if retry_count > 0:
                print()
        except:
            print(f'OCC error in node {k}. Node list:')
            print(nodelist)
            raise
        if occ_nsides is None:
            occ_nsides = node_nsides
        processed_nodes.append((node_verts, node_faces, node_edverts, nt, nodelist, ))
        total_verts += len(node_verts) * len(nodelist)
        total_faces += len(node_faces) * len(nodelist)

    nsides = occ_nsides
    print('Nsides:', nsides)
    newverts = np.empty((total_verts, 3), dtype=np.float32)
    newfaces = np.empty((total_faces, 3), dtype=np.int32)

    vert_count = 0
    face_count = 0
    edgefaces = np.zeros((nedges, 2), dtype=np.int32)
    ed_nd = np.zeros((nedges, 2, 2), dtype=np.int32)
    for k, (node_verts, node_faces, node_edverts, nt, nodelist) in enumerate(processed_nodes):
        valency = len(node_edverts)
        verts_per_node = len(node_verts)
        faces_per_node = len(node_faces)
        ed_pos = np.arange(valency, dtype=np.int32)
        if valency:
            for n in nodelist:
                # Translate the vertices of each node type to the actual position of the nodes and
                # insert them in newverts
                if verts_per_node:
                    newverts[vert_count:vert_count + verts_per_node, :] = node_verts + verts[n,:]
                # Insert the node type faces in newfaces for each of the node instances
                if faces_per_node:
                    newfaces[face_count:face_count + faces_per_node, :] = node_faces + vert_count
                edgefaces[nodes[n][:,0], 1 - nodes[n][:,1]] = vert_count
                ed_nd[nodes[n][:,0],0, 1 - nodes[n][:,1]] = k
                ed_nd[nodes[n][:,0],1, 1 - nodes[n][:,1]] = ed_pos
                vert_count += verts_per_node
                face_count += faces_per_node
        else:
            print('Edgeless node found. Node list:', nodelist)

    ed_nd_dict = {}
    for k in range(nedges):
        if k in discarded_edges:
            continue
        node_a = tuple(ed_nd[k, :, 0])
        node_b = tuple(ed_nd[k, :, 1])
        edkey = tuple(sorted([node_a, node_b]))
        edir = int(node_a == edkey[0])
        ed_nd_dict.setdefault(edkey, []).append([k, edir])

    print('Edge node types:', len(ed_nd_dict))
    
    ed_nd_faces = {}
    ed_nd_nfaces = 0
    for ed_ty in ed_nd_dict.keys():
        (na, fa), (nb, fb) = ed_ty
        node_verts, node_faces, node_edverts, nt, nodelist = processed_nodes[na]
        ed_dir = np.array(listdir[nt[fa]], dtype=np.float32) / quantization
        nedidxa = node_edverts[fa]
        nva = node_verts[nedidxa, :]
        node_verts, node_faces, node_edverts, nt, nodelist = processed_nodes[nb]
        nedidxb = node_edverts[fb]
        nvb = node_verts[nedidxb, :]
        rotmat = make_rotmatrix(ed_dir)
        rotmat = rotmat.T[:,:2].copy()
        proj_verts = nva @ rotmat
        ang_a = np.arctan2(proj_verts[:,1], proj_verts[:,0])
        proj_verts = nvb @ rotmat
        ang_b = np.arctan2(proj_verts[:,1], proj_verts[:,0])
        ed_fa, face_masq = ring_connect_faces(ang_a, ang_b)
        other_masq = np.logical_not(face_masq)
        ed_fanew = np.zeros(ed_fa.shape, dtype=np.int32)
        ed_fanew[face_masq] = nedidxa[ed_fa[face_masq]]
        ed_fanew[other_masq] = nedidxb[ed_fa[other_masq]]
        ed_nd_faces[ed_ty] = [ed_fanew, face_masq]
        ed_nd_nfaces += len(ed_nd_dict[ed_ty]) * len(ed_fa)

    newfaces.resize((total_faces + ed_nd_nfaces, 3))
    print('Nb. of edge node faces:', ed_nd_nfaces)
        

    processed_faces = 0
    for ed_ty in ed_nd_dict.keys():
        edfaces, face_masq = ed_nd_faces[ed_ty]
        other_facemasq = np.logical_not(face_masq)
        ed_nfaces = len(edfaces)
        faceoffsets = np.zeros(edfaces.shape, dtype=np.int32)
        for edidx, edir in ed_nd_dict[ed_ty]:
            faceoffsets[face_masq] = edgefaces[edidx, 1-edir]
            faceoffsets[other_facemasq] = edgefaces[edidx, edir]
            newfaces[total_faces + processed_faces : total_faces + processed_faces + ed_nfaces,:] = edfaces + faceoffsets
            processed_faces += ed_nfaces
        
        

    mesh.from_pydata(newverts, [], list(newfaces))
    newobj  = bpy.data.objects.new(name, mesh)

    # Place the lattice at the same position and attitude of original object       
    newobj.matrix_world = target.matrix_world.copy()


    # Link object to scene
    bpy.context.scene.collection.objects.link(newobj)

    

def createMetaball(target,  r=0.5, p=0.9):
    if target.type != 'MESH':
        return
    
    name = target.name + "Rod"
    print("Creating", name)

    metaball = bpy.data.metaballs.new(name+'Meta')
    obj = bpy.data.objects.new(name, metaball)
    bpy.context.scene.collection.objects.link(obj)
    stiffness = 10.0
    
    # clip p
    p = min(p, 0.99);
    p = max(p, 0.01)

    # Compute the threashold so that mesh radius is 
    # exacltly r
    metaball.threshold = (1 - p ** 2) ** 3 * stiffness
    
    radius = r / p
    
    vec_x = Vector((1.0, 0.0, 0.0))

    for v in target.data.vertices:
        location = v.co.copy()

        element = metaball.elements.new()
        element.co = location
        element.radius = radius
        element.stiffness = stiffness

    for e in target.data.edges:
        v0 = target.data.vertices[e.vertices[0]].co
        v1 = target.data.vertices[e.vertices[1]].co
        edge = v1 - v0
        edge_len = edge.length

        if edge_len > r:
            edge_middle = (v1 + v0) / 2.0
            element = metaball.elements.new()
            element.co = edge_middle
            element.radius = radius
            element.stiffness = stiffness

            if edge_len > 2 * r:
                element.type = 'CAPSULE'
                element.size_x = (edge_len - 2 * r) / 2.0
                angle = edge.angle(vec_x)
                if abs(angle) > 1e-3:
                    axis = edge.cross(vec_x).normalized()
                    rot = Quaternion(axis, -angle).normalized()
                    element.rotation = (rot.w, rot.x, rot.y, rot.z)
                    
                
    # Place the lattice at the same position and attitude of original object       
    obj.matrix_world = target.matrix_world.copy()

    return obj


class ObjectMetaLattice(bpy.types.Operator):
    """Metaball Lattice"""
    bl_idname = "object.metalattice"
    bl_label = "Meta Lattice"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        return context.active_object is not None and context.active_object.type == 'MESH'


    def execute(self, context):
        lattice_mesh = context.scene.lattice_mesh
        createMetaball(context.active_object, lattice_mesh.radius, lattice_mesh.r_ratio)
        return {'FINISHED'}


class ObjectOCCLattice(bpy.types.Operator):
    """Metaball Lattice"""
    bl_idname = "object.occlattice"
    bl_label = "Lattice Volume (OCC)"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        return have_occ and context.active_object is not None and context.active_object.type == 'MESH'


    def execute(self, context):
        lattice_mesh = context.scene.lattice_mesh
        createMesh(context.active_object, lattice_mesh.radius, lattice_mesh.lin_deflect, lattice_mesh.ang_deflect)
        return {'FINISHED'}



class ObjectCreateMeshLattice(bpy.types.Operator):
    """Create Mesh Lattice"""
    bl_idname = "object.create_meshlattice"
    bl_label = "Create Mesh Lattice"
    bl_options = {'REGISTER', 'UNDO'}
    


    @classmethod
    def poll(cls, context):
        return context.mode == 'OBJECT'


    def execute(self, context):
        lattice_mesh = context.scene.lattice_mesh
        name = "Lattice"
        nelems = np.array(lattice_mesh.nelems)
        corner = np.array(context.scene.cursor.location).reshape((1,3))
        v, e = make_nplattice(Cells[lattice_mesh.celltype], lattice_mesh.cellsize, nelems, corner)
        verts, edges = nplattice2meshdata(v, e)
        bm, mesh = make_meshlattice(name, verts, edges, lattice_mesh.cellsize)
        bm.to_mesh(mesh)
        bm.free()
        
        newobj  = bpy.data.objects.new(name, mesh)

        # Link object to scene
        bpy.context.scene.collection.objects.link(newobj)

        return {'FINISHED'}


class ObjectFillMeshLattice(bpy.types.Operator):
    """Fill Mesh Lattice"""
    bl_idname = "object.fill_meshlattice"
    bl_label = "Fill Mesh Lattice"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        return context.active_object is not None and context.active_object.type == 'MESH'


    def execute(self, context):
        obj = context.active_object
        name = obj.name + "Lattice"
        bb = np.array([[x for x in y] for y in obj.bound_box], dtype=np.float32)
        minp = bb.min(axis=0)
        maxp = bb.max(axis=0)
        dims = maxp - minp
        lattice_mesh = context.scene.lattice_mesh
        nelems = np.ceil((dims + lattice_mesh.cellsize)/ lattice_mesh.cellsize).astype(np.int32)
        griddims = nelems.astype(np.float32) * lattice_mesh.cellsize
        corner = ((maxp + minp) / 2.0) - (griddims /2.0)
        corner = corner.reshape((1,3))

        v, e = make_nplattice(Cells[lattice_mesh.celltype], lattice_mesh.cellsize, nelems, corner)
        verts, edges = nplattice2meshdata(v, e)
        bm, mesh = make_meshlattice(name, verts, edges, lattice_mesh.cellsize)
        depsgraph = context.evaluated_depsgraph_get()
        intersect(bm, obj, depsgraph)
        bm.to_mesh(mesh)
        bm.free()
        
        newobj  = bpy.data.objects.new(name, mesh)
        newobj.matrix_world = obj.matrix_world.copy()

        # Link object to scene
        bpy.context.scene.collection.objects.link(newobj)

        return {'FINISHED'}


class ObjectFillVoronoiLattice(bpy.types.Operator):
    """Fill Voronoi Lattice"""
    bl_idname = "object.fill_voronoilattice"
    bl_label = "Fill Voronoi Lattice"
    bl_options = {'REGISTER', 'UNDO'}
    
    @classmethod
    def poll(cls, context):
        return have_scipy and context.active_object is not None and context.active_object.type == 'MESH'

    def execute(self, context):
        obj = context.active_object
        name = obj.name + "Voronoi"
        bb = np.array([[x for x in y] for y in obj.bound_box], dtype=np.float32)
        minp = bb.min(axis=0)
        maxp = bb.max(axis=0)
        dims = maxp - minp
        corner = minp
        corner = corner.reshape((1,3))

        lattice_mesh = context.scene.lattice_mesh
        v, e = make_npvoronoi(lattice_mesh.charac_length, dims, corner)
        verts, edges = nplattice2meshdata(v, e)
        bm, mesh = make_meshlattice(name, verts, edges, lattice_mesh.charac_length)
        depsgraph = context.evaluated_depsgraph_get()
        intersect(bm, obj, depsgraph)
        bm.to_mesh(mesh)
        bm.free()
        
        newobj  = bpy.data.objects.new(name, mesh)
        newobj.matrix_world = obj.matrix_world.copy()

        # Link object to scene
        bpy.context.scene.collection.objects.link(newobj)

        return {'FINISHED'}




